@@ -1,3 +1,4 @@
+#include <cosma/math_utils.hpp>
 #include <cosma/local_multiply.hpp>
 #include <cosma/multiply.hpp>
 #include <cosma/profiler.hpp>
@@ -6,6 +7,14 @@
 
 #include <complex>
 
+#if defined(COSMA_HAVE_GPU) && defined(COSMA_WITH_NCCL)
+#include <cosma/gpu/nccl_utils.hpp>
+#endif
+
+#if defined(COSMA_HAVE_GPU) && defined(COSMA_WITH_GPU_AWARE_MPI)
+#include <cosma/gpu/gpu_aware_mpi_utils.hpp>
+#endif
+
 namespace cosma {
 template <typename Scalar>
 void multiply(cosma_context<Scalar> *ctx,
@@ -18,7 +27,7 @@ void multiply(cosma_context<Scalar> *ctx,
               Interval &P,
               size_t step,
               const Strategy &strategy,
-              communicator &comm,
+              communicator *comm,
               Scalar alpha,
               Scalar beta);
 
@@ -33,7 +42,7 @@ void sequential(cosma_context<Scalar> *ctx,
                 Interval &P,
                 size_t step,
                 const Strategy &strategy,
-                communicator &comm,
+                communicator *comm,
                 Scalar alpha,
                 Scalar beta);
 
@@ -48,7 +57,7 @@ void parallel(cosma_context<Scalar> *ctx,
               Interval &P,
               size_t step,
               const Strategy &strategy,
-              communicator &comm,
+              communicator *comm,
               Scalar alpha,
               Scalar beta);
 
@@ -144,12 +153,22 @@ void multiply_using_layout(cosma_context<T> *ctx,
     CosmaMatrix<T> B_cosma(ctx, std::move(mapper_b), rank_permutation[rank]);
     CosmaMatrix<T> C_cosma(ctx, std::move(mapper_c), rank_permutation[rank]);
 
-    // avoid resizing of buffer by reserving immediately the total required memory
-    ctx->get_memory_pool().reserve(
-            A_cosma.total_required_memory()
-          + B_cosma.total_required_memory()
-          + C_cosma.total_required_memory()
-    );
+    // avoid resizing the buffer by reserving immediately the total required memory
+    // collect sizes of all buffers that are going to be allocated for each matrix
+    auto A_buffers = A_cosma.required_memory();
+    auto B_buffers = B_cosma.required_memory();
+    auto C_buffers = C_cosma.required_memory();
+    std::vector<std::size_t> buffer_sizes;
+    int n_buffers = A_buffers.size() + B_buffers.size() + C_buffers.size();
+    if (n_buffers > 0) {
+        buffer_sizes.reserve(n_buffers);
+        std::copy(A_buffers.begin(), A_buffers.end(), std::back_inserter(buffer_sizes));
+        std::copy(B_buffers.begin(), B_buffers.end(), std::back_inserter(buffer_sizes));
+        std::copy(C_buffers.begin(), C_buffers.end(), std::back_inserter(buffer_sizes));
+
+        // allocate all buffers in the memory pool
+        get_context_instance<T>()->get_memory_pool().reserve(buffer_sizes);
+    }
 
     // get abstract layouts for COSMA layout
     auto cosma_layout_a = A_cosma.get_grid_layout();
@@ -231,6 +250,12 @@ void multiply(cosma_context<Scalar> *ctx,
         return;
     }
 
+    // register reusable objects in the context
+    ctx->register_state(comm, strategy);
+    if (comm == MPI_COMM_NULL || ctx->get_cosma_comm()->is_idle()) {
+	return;
+    }
+
     Interval mi = Interval(0, strategy.m - 1);
     Interval ni = Interval(0, strategy.n - 1);
     Interval ki = Interval(0, strategy.k - 1);
@@ -257,28 +282,19 @@ void multiply(cosma_context<Scalar> *ctx,
     assert(matrixB.rank() == matrixC.rank());
     PL();
 
-    PE(preprocessing_communicators);
-    communicator cosma_comm = communicator(&strategy, comm);
-    PL();
-
-    if (!cosma_comm.is_idle()) {
-        // register strategy in the context
-        ctx->register_state(cosma_comm.rank(), strategy);
-
-        multiply(ctx,
-                 matrixA,
-                 matrixB,
-                 matrixC,
-                 mi,
-                 ni,
-                 ki,
-                 Pi,
-                 0,
-                 strategy,
-                 cosma_comm,
-                 alpha,
-                 beta);
-    }
+    multiply(ctx,
+    	 matrixA,
+    	 matrixB,
+    	 matrixC,
+    	 mi,
+    	 ni,
+    	 ki,
+    	 Pi,
+    	 0,
+    	 strategy,
+    	 ctx->get_cosma_comm(),
+    	 alpha,
+    	 beta);
 
     // deallocate buffers used for communication
     // since its a stack allocator, we deallocate
@@ -289,7 +305,7 @@ void multiply(cosma_context<Scalar> *ctx,
     matrixA.free_communication_buffers();
     PL();
 
-    if (cosma_comm.rank() == 0) {
+    if (ctx->get_cosma_comm()->rank() == 0) {
         PP();
     }
 }
@@ -305,7 +321,7 @@ void multiply(cosma_context<Scalar> *ctx,
               Interval &P,
               size_t step,
               const Strategy &strategy,
-              communicator &comm,
+              communicator *comm,
               Scalar alpha,
               Scalar beta) {
     PE(multiply_other);
@@ -342,12 +358,27 @@ void multiply(cosma_context<Scalar> *ctx,
 
     // This iterates over the skipped buckets and sums up their sizes,
     // and increases the pointer of the current matrix for the offset
-    int offsetA = matrixA.shift(bucketA[comm.relative_rank(P)]);
-    int offsetB = matrixB.shift(bucketB[comm.relative_rank(P)]);
-    int offsetC = matrixC.shift(bucketC[comm.relative_rank(P)]);
+    int offsetA = matrixA.shift(bucketA[comm->relative_rank(P)]);
+    int offsetB = matrixB.shift(bucketB[comm->relative_rank(P)]);
+    int offsetC = matrixC.shift(bucketC[comm->relative_rank(P)]);
     PL();
 
     if (strategy.final_step(step) || strategy.empty()) {
+        bool copy_c_back = true;
+        bool nccl_enabled = false;
+        bool gpu_aware_mpi_enabled = false;
+
+#ifdef COSMA_WITH_NCCL
+        nccl_enabled = true;
+#endif
+#ifdef COSMA_WITH_GPU_AWARE_MPI
+        gpu_aware_mpi_enabled = true;
+#endif
+
+        if (gpu_aware_mpi_enabled || nccl_enabled) {
+            copy_c_back = !(step > 0 && strategy.parallel_step(step-1) && strategy.split_k(step-1));
+        }
+
         local_multiply(ctx,
                        matrixA.current_matrix(),
                        matrixB.current_matrix(),
@@ -356,11 +387,12 @@ void multiply(cosma_context<Scalar> *ctx,
                        n.length(),
                        k.length(),
                        alpha,
-                       beta);
+                       beta,
+                       copy_c_back);
     } else {
         if (strategy.parallel_step(step)) {
             if (strategy.should_overlap_comm_and_comp(step)) {
-                comm.overlap_comm_and_comp(ctx,
+                comm->overlap_comm_and_comp(ctx,
                                            matrixA,
                                            matrixB,
                                            matrixC,
@@ -433,7 +465,7 @@ void sequential(cosma_context<Scalar> *ctx,
                 Interval &P,
                 size_t step,
                 const Strategy &strategy,
-                communicator &comm,
+                communicator *comm,
                 Scalar alpha,
                 Scalar beta) {
     // split the dimension but not the processors, all P processors are taking
@@ -620,7 +652,7 @@ void parallel(cosma_context<Scalar> *ctx,
               Interval &P,
               size_t step,
               const Strategy &strategy,
-              communicator &comm,
+              communicator *comm,
               Scalar alpha,
               Scalar beta) {
     PE(multiply_other);
@@ -629,7 +661,7 @@ void parallel(cosma_context<Scalar> *ctx,
     int divisor_n = strategy.divisor_n(step);
     int divisor_k = strategy.divisor_k(step);
     // processor subinterval which the current rank belongs to
-    int partition_idx = P.subinterval_index(divisor, comm.rank());
+    int partition_idx = P.subinterval_index(divisor, comm->rank());
     Interval newP = P.subinterval(divisor, partition_idx);
     // intervals of M, N and K that the current rank is in charge of,
     // together with other ranks from its group.
@@ -699,7 +731,7 @@ void parallel(cosma_context<Scalar> *ctx,
     // this is the sum of sizes of all the buckets after expansion
     // that the current rank will own.
     // which is also the size of the matrix after expansion
-    int new_size = total_after_expansion[comm.relative_rank(newP)];
+    int new_size = total_after_expansion[comm->relative_rank(newP)];
 
     int buffer_idx = expanded_mat.buffer_index();
     expanded_mat.advance_buffer();
@@ -717,7 +749,29 @@ void parallel(cosma_context<Scalar> *ctx,
     // should own exactly the same data in the expanded matrix.
     if (strategy.split_m(step) || strategy.split_n(step)) {
         // copy the matrix that wasn't divided in this step
-        comm.copy(P,
+#ifdef COSMA_WITH_NCCL
+        cosma::gpu::nccl_copy(ctx,
+                              P,
+                              original_matrix,
+                              expanded_matrix,
+                              reshuffle_buffer,
+                              size_before_expansion,
+                              total_before_expansion,
+                              new_size,
+                              step);
+#elif COSMA_WITH_GPU_AWARE_MPI
+        cosma::gpu::gpu_aware_mpi_copy(
+                              ctx,
+                              P,
+                              original_matrix,
+                              expanded_matrix,
+                              reshuffle_buffer,
+                              size_before_expansion,
+                              total_before_expansion,
+                              new_size,
+                              step);
+#else
+        comm->copy(P,
                   original_matrix,
                   expanded_matrix,
                   reshuffle_buffer,
@@ -725,6 +779,7 @@ void parallel(cosma_context<Scalar> *ctx,
                   total_before_expansion,
                   new_size,
                   step);
+#endif
     }
 
     // if division by k, and we are in the branch where beta > 0, then
@@ -825,10 +880,10 @@ void parallel(cosma_context<Scalar> *ctx,
              new_beta);
 
 #ifdef DEBUG
-    std::cout << "rank = " << comm.rank() << ", label = " << expanded_mat.label() << std::endl;
-    if (comm.rank() == 0 && expanded_mat.label() == 'C') {
+    std::cout << "rank = " << comm->rank() << ", label = " << expanded_mat.label() << std::endl;
+    if (comm->rank() == 0 && expanded_mat.label() == 'C') {
         std::cout << "expanded matrix after multiply: " << std::endl;
-        int local_size = size_before_expansion[comm.rank() - P.first()][0];
+        int local_size = size_before_expansion[comm->rank() - P.first()][0];
         for (int i = 0; i < local_size; ++i) {
             std::cout << *(expanded_mat.current_matrix() + i) << ", ";
         }
@@ -850,7 +905,39 @@ void parallel(cosma_context<Scalar> *ctx,
     // if division by k do additional reduction of C
     if (strategy.split_k(step)) {
         Scalar *reduce_buffer = expanded_mat.reduce_buffer_ptr();
-        comm.reduce(P,
+#ifdef COSMA_WITH_NCCL
+        bool copy_c_back = !strategy.final_step(step+1);
+        cosma::gpu::nccl_reduce(ctx,
+                                P,
+                                expanded_matrix,
+                                original_matrix,
+                                reshuffle_buffer,
+                                reduce_buffer,
+                                size_before_expansion,
+                                total_before_expansion,
+                                size_after_expansion,
+                                total_after_expansion,
+                                beta,
+                                step,
+                                copy_c_back);
+#elif COSMA_WITH_GPU_AWARE_MPI
+        bool copy_c_back = !strategy.final_step(step+1);
+        cosma::gpu::gpu_aware_mpi_reduce(
+                                ctx,
+                                P,
+                                expanded_matrix,
+                                original_matrix,
+                                reshuffle_buffer,
+                                reduce_buffer,
+                                size_before_expansion,
+                                total_before_expansion,
+                                size_after_expansion,
+                                total_after_expansion,
+                                beta,
+                                step,
+                                copy_c_back);
+#else
+        comm->reduce(P,
                     expanded_matrix,
                     original_matrix,
                     reshuffle_buffer,
@@ -862,6 +949,7 @@ void parallel(cosma_context<Scalar> *ctx,
                     alpha,
                     beta,
                     step);
+#endif
     }
 
     PE(multiply_other);
