@@ -482,6 +482,7 @@ void overlap_m_split(bool use_busy_waiting,
     matrixC.set_current_matrix(pointer_c);
 
     PL();
+    bool copy_c_back = true;
     local_multiply(ctx,
                    matrixA.current_matrix(),
                    matrixB.current_matrix(),
@@ -490,7 +491,8 @@ void overlap_m_split(bool use_busy_waiting,
                    n.subinterval(divisor, gp).length(),
                    k.length(),
                    alpha,
-                   beta);
+                   beta,
+                   copy_c_back);
     PE(multiply_communication_other);
 
     int dist = 1;
@@ -515,7 +517,8 @@ void overlap_m_split(bool use_busy_waiting,
                            n.subinterval(divisor, idx).length(),
                            k.length(),
                            alpha,
-                           beta);
+                           beta,
+                           copy_c_back);
             PE(multiply_communication_copy);
             ready--;
             dist++;
@@ -625,6 +628,7 @@ void overlap_n_split(bool use_busy_waiting,
 
             Scalar new_beta = dist == 0 ? beta : Scalar(1);
             PL();
+            bool copy_c_back = true;
             local_multiply(ctx,
                            matrixA.current_matrix(),
                            matrixB.current_matrix(),
@@ -633,7 +637,8 @@ void overlap_n_split(bool use_busy_waiting,
                            newn.length(),
                            k.subinterval(divisor, idx).length(),
                            alpha,
-                           new_beta);
+                           new_beta,
+                           copy_c_back);
             PE(multiply_communication_other);
 
             dist++;
@@ -750,6 +755,7 @@ void compute(cosma_context<Scalar> *ctx,
     // C.set_current_matrix(c);
 
     PL();
+    bool copy_c_back = true;
     local_multiply(ctx,
                    A.current_matrix(),
                    B.current_matrix(),
@@ -758,7 +764,8 @@ void compute(cosma_context<Scalar> *ctx,
                    n_length,
                    k.length(),
                    alpha,
-                   beta);
+                   beta,
+                   copy_c_back);
     PE(multiply_communication_other);
 }
 
@@ -1014,7 +1021,7 @@ template <typename Scalar>
 void overlap_comm_and_comp(cosma_context<Scalar> *ctx,
                            MPI_Comm comm,
                            int rank,
-                           const Strategy *strategy,
+                           const Strategy strategy,
                            CosmaMatrix<Scalar> &matrixA,
                            CosmaMatrix<Scalar> &matrixB,
                            CosmaMatrix<Scalar> &matrixC,
@@ -1025,9 +1032,9 @@ void overlap_comm_and_comp(cosma_context<Scalar> *ctx,
                            size_t step,
                            Scalar alpha,
                            Scalar beta) {
-    bool use_busy_waiting = strategy->use_busy_waiting;
-    int divisor = strategy->divisor(step);
-    if (strategy->split_m(step)) {
+    bool use_busy_waiting = strategy.use_busy_waiting;
+    int divisor = strategy.divisor(step);
+    if (strategy.split_m(step)) {
         overlap_m_split(use_busy_waiting,
                         ctx,
                         comm,
@@ -1042,7 +1049,7 @@ void overlap_comm_and_comp(cosma_context<Scalar> *ctx,
                         P,
                         alpha,
                         beta);
-    } else if (strategy->split_n(step)) {
+    } else if (strategy.split_n(step)) {
         overlap_n_split(use_busy_waiting,
                         ctx,
                         comm,
@@ -1077,7 +1084,7 @@ void overlap_comm_and_comp(cosma_context<Scalar> *ctx,
 template void overlap_comm_and_comp<float>(cosma_context<float> *ctx,
                                            MPI_Comm comm,
                                            int rank,
-                                           const Strategy *strategy,
+                                           const Strategy strategy,
                                            CosmaMatrix<float> &matrixA,
                                            CosmaMatrix<float> &matrixB,
                                            CosmaMatrix<float> &matrixC,
@@ -1092,7 +1099,7 @@ template void overlap_comm_and_comp<float>(cosma_context<float> *ctx,
 template void overlap_comm_and_comp<double>(cosma_context<double> *ctx,
                                             MPI_Comm comm,
                                             int rank,
-                                            const Strategy *strategy,
+                                            const Strategy strategy,
                                             CosmaMatrix<double> &matrixA,
                                             CosmaMatrix<double> &matrixB,
                                             CosmaMatrix<double> &matrixC,
@@ -1108,7 +1115,7 @@ template void overlap_comm_and_comp<std::complex<float>>(
     cosma_context<std::complex<float>> *ctx,
     MPI_Comm comm,
     int rank,
-    const Strategy *strategy,
+    const Strategy strategy,
     CosmaMatrix<std::complex<float>> &matrixA,
     CosmaMatrix<std::complex<float>> &matrixB,
     CosmaMatrix<std::complex<float>> &matrixC,
@@ -1124,7 +1131,7 @@ template void overlap_comm_and_comp<std::complex<double>>(
     cosma_context<std::complex<double>> *ctx,
     MPI_Comm comm,
     int rank,
-    const Strategy *strategy,
+    const Strategy strategy,
     CosmaMatrix<std::complex<double>> &matrixA,
     CosmaMatrix<std::complex<double>> &matrixB,
     CosmaMatrix<std::complex<double>> &matrixC,
